#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Trajectory Reader and Writer superclasses."""

# ============================================================================
# IMPORTS
# ============================================================================

from ..core import AtomicSystem


# ============================================================================
# CLASSES
# ============================================================================


class TrajectoryReader:
    """Class to read trajectory files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories in xyz format are

    ftype : str
        different type of files depending on the child class.
    """

    def __init__(self, filename, ftype):
        self.filename = filename
        self.ftype = ftype

        self.frame = AtomicSystem()

    def __enter__(self):
        """Use the open() method."""
        self.file_traj_ = open(self.filename, "r")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Use the close() method."""
        self.file_traj_.close()

    def read_frame(self):
        """Read the actual frame of the file."""
        raise NotImplementedError("Implemented in child classes.")


class TrajectoryWriter:
    """Class to write trajectory files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories in xyz format are going to
        be written

    ftype : str
        different type of files depending on the child class.
    """

    def __init__(self, filename, ftype):
        self.filename = filename
        self.ftype = ftype

    def __enter__(self):
        """Use the open() method."""
        self.file_traj_ = open(self.filename, "w")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Use the close() method."""
        self.file_traj_.close()

    def write_frame(self):
        """Write the actual frame on the file."""
        raise NotImplementedError("Implemented in child classes.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Distances with Periodic Boundary Conditions."""

# ============================================================================
# IMPORTS
# ============================================================================

import ctypes as ct
import os
import pathlib
import sysconfig

import numpy as np

# ============================================================================
# CONSTANTS
# ============================================================================

PATH = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))


# ============================================================================
# FUNCTIONS
# ============================================================================


def pbc_distances(frame_c, frame_i, type_c, type_i):
    """Periodic boundary conditions distances.

    A function to compute the distances between a central and an interact
    group of atoms with the minimum image convention, in which each individual
    particle inteacts with the closest image of the copies of the system.

    Parameters
    ----------
    frame_c : `exma.core.AtomicSystem`
        which have the atomic configurations of the central atoms and must
        include `box` not None.

    frame_i : `exma.core.AtomicSystem`
        which have the atomic configurations of the interact atoms and must
        include `box` not None.

    type_c : int or str
        type of central atoms

    type_i : int or str
        type of interacting atoms

    Returns
    -------
    np.array
        array with a vector where the first natoms_i components are the
        distances of the interacting atoms to the first central one, then the
        second natoms_i to the second central atom, etc.
    """
    lib_pbc_distances = ct.CDLL(
        str(PATH / "lib" / "lib_pbc_distances")
        + sysconfig.get_config_var("EXT_SUFFIX")
    )
    distance_matrix_c = lib_pbc_distances.distance_matrix
    distance_matrix_c.argtypes = [
        ct.c_int,
        ct.c_int,
        ct.c_void_p,
        ct.c_void_p,
        ct.c_void_p,
        ct.c_void_p,
    ]

    mask_c = frame_c._mask_type(type_c)
    mask_i = frame_i._mask_type(type_i)
    natoms_c = frame_c._natoms_type(mask_c)
    natoms_i = frame_i._natoms_type(mask_i)

    box = frame_i.box
    xc, yc, zc = frame_c.x[mask_c], frame_c.y[mask_c], frame_c.z[mask_c]
    xi, yi, zi = frame_i.x[mask_i], frame_i.y[mask_i], frame_i.z[mask_i]

    distrix = np.zeros(natoms_c * natoms_i, dtype=np.float32)

    # prepare data for C code
    box = np.asarray(box, dtype=np.float32)
    box_c = box.ctypes.data_as(ct.POINTER(ct.c_void_p))

    xcentral_c = np.concatenate((xc, yc, zc)).astype(np.float32)
    xcentral_c = xcentral_c.ctypes.data_as(ct.POINTER(ct.c_void_p))

    xinteract_c = np.concatenate((xi, yi, zi)).astype(np.float32)
    xinteract_c = xinteract_c.ctypes.data_as(ct.POINTER(ct.c_void_p))

    distrix_c = distrix.ctypes.data_as(ct.POINTER(ct.c_void_p))

    # calculates the distance matrix between interact and central atoms
    distance_matrix_c(
        natoms_c, natoms_i, box_c, xcentral_c, xinteract_c, distrix_c
    )

    return distrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Implementations of clustering."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import scipy.integrate

import sklearn.cluster

from ._mdobservable import MDObservable
from .distances import pbc_distances


# ============================================================================
# CLASSES
# ============================================================================


class EffectiveNeighbors(MDObservable):
    """Emipirical way to describe charge transfer and coordination in solids.

    The empirical effective coordination model [2]_, used to calculate the
    effective neighbors, assumes that the interact atoms donate more of its
    electron to the closest central atoms. Then, fractions of the interact
    atom can be assigned to the various central neighbors atoms [3]_.

    Parameters
    ----------
    ftraj : str
        the string corresponding with the filename with the molecular
        dynamics trajectory

    type_c : int or str
        type of central atoms

    type_i : int or str
        type of interacting atoms

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames

    References
    ----------
    .. [2] Hoppe, R., Voigt, S., Glaum, H., Kissel, J., Müller, H.P. and
       Bernet, K., 1988. A new route to charge distributions in ionic solids.
       `Journal of the Less Common Metals`, 156(1-2), pp.105-122.
    .. [3] Chevrier, V.L. and Dahn, J.R., 2010. First principles studies of
       disordered lithiated silicon. `Journal of the Electrochemical Society`,
       157(4), p.A392.

    """

    def __init__(self, ftraj, type_c, type_i, start=0, stop=-1, step=1):
        super().__init__(ftraj, start, stop, step)

        self.type_c = type_c
        self.type_i = type_i

    def _local_configure(self, frame):
        """Configure the Effective Neighbors calculus."""
        self.natoms_c_ = frame._natoms_type(frame._mask_type(self.type_c))
        self.natoms_i_ = frame._natoms_type(frame._mask_type(self.type_i))

        self.counter_ = 0
        self.effnei_ = np.zeros(self.natoms_c_, dtype=np.float32)

    def _accumulate(self, frame):
        """Obtain the efective (interact) neighbors of the actual frame.

        Parameters
        ----------
        frame : `exma.core.AtomicSystem`
            with the information of the atomic system including the `box`
        """
        distrix = pbc_distances(frame, frame, self.type_c, self.type_i)

        # calculate the weigth of the ith neighbor of the interact atom
        bondmin = np.min(distrix)
        matrix_a = np.exp(1.0 - np.power(distrix / bondmin, 6))
        bondavg = np.sum(distrix * matrix_a) / np.sum(matrix_a)

        weitrix = np.exp(1.0 - np.power(distrix / bondavg, 6))

        # reshape the weight matrix and transpose it to obtain an interact
        # atom in every row and normalize their weights
        weitrix = np.reshape(weitrix, (self.natoms_c_, self.natoms_i_)).T
        weitrix = [
            weitrix[i] / np.sum(weitrix[i]) for i in range(self.natoms_i_)
        ]

        # the matrix is transpose again so now we have central atoms in each
        # row an each fraction of every interact neighbor is added to obtain
        # the effective (interact) neighbor
        weitrix = np.transpose(weitrix)
        effnei = [np.sum(weitrix[i]) for i in range(self.natoms_c_)]

        self.counter_ += 1
        self.effnei_ = self.effnei_ + np.array(effnei, dtype=np.float32)

    def _end(self):
        """Complete the calculation by normalizing the data."""
        return self.effnei_ / self.counter_

    def calculate(self, box=None):
        """Calculate the Effective Neighbors.

        Parameters
        ----------
        box : np.array, default=None
            the lenght of the box in each x, y, z direction, required when
            the trajectory is in an xyz file.

        Returns
        -------
        np.array
            effective (interact) neighbor of the central atoms in the same
            order that are in the positions vector
        """
        super()._calculate(box)

        return self._end()


class DBSCAN:
    """DBSCAN clustering using PBC for the distance matrix calculations.

    Parameters
    ----------
    type_c : int or str
        type of central atoms

    type_i : int or str
        type of interacting atoms

    eps : float
        a cutoff radius at which an atom is no longer considered part of the
        cluster.

    min_samples : int, default=2
        number of atoms required to form a cluster.

    Notes
    -----
    If the system is not under periodic boundary conditions, then
    everything can be calculated from `sklearn.cluster.DBSCAN` using
    `metrics="euclidean"`.
    """

    def __init__(self, type_c, type_i, eps, min_samples=2):
        self.type_c = type_c
        self.type_i = type_i
        self.eps = eps
        self.min_samples = min_samples

    def of_this_frame(self, frame, **kwargs):
        """Obtain the labels of the DBSCAN clustering the actual frame.

        Parameters
        ----------
        frame : `exma.core.AtomicSystem`
            with the information of the atomic system including the `box`

        **kwargs
            additional keyword arguments that are passed and are documented
            in `sklearn.cluster.DBSCAN`

        Returns
        -------
        id_cluster : np.array
            as explained in `sklearn.cluster.DBSCAN`. It contains the id
            number of the cluster to which belongs the corresponding atom
            (the array is sorted). A value of -1 means that the atom is
            isolated.
        """
        self.natoms_c_ = frame._natoms_type(frame._mask_type(self.type_c))
        self.natoms_i_ = frame._natoms_type(frame._mask_type(self.type_i))

        distrix = pbc_distances(frame, frame, self.type_c, self.type_i)

        distrix = distrix.reshape((self.natoms_c_, self.natoms_i_))
        db = sklearn.cluster.DBSCAN(
            eps=self.eps,
            min_samples=self.min_samples,
            metric="precomputed",
            **kwargs,
        ).fit(distrix)

        self.id_cluster_ = np.asarray(db.labels_, dtype=np.intc)

        return self.id_cluster_

    def characterize(self):
        """Characterization in number of clusters and isolated.

        Returns
        -------
        tuple
            with the number of clusters and the number of isolated atoms.
        """
        isolated = np.count_nonzero(self.id_cluster_ == -1)

        uniques = np.unique(self.id_cluster_).size
        clusters = uniques if isolated == 0 else uniques - 1

        return clusters, isolated


# ============================================================================
# FUNCTIONS
# ============================================================================


def sro(rdf_x, rdf_y, rcut, **kwargs):
    """Short range order parameter.

    To characterize the short range ordering of amorphous structures, defined
    in this work [4]_, using the itegration of the radial distribution
    function for diatomic systems.

    This parameter indicates complete randomness if it is equal to zero,
    preference for unlike neighbors if it is less than zero, and preference
    for similar neighbors (clustering) if is greater than zero.

    Parameters
    ----------
    rdf_x : np.array
        x of the radial distribution function

    rdf_y : np.array
        y of the radial distribution function

    rcut : float
        cutoff radius

    **kwargs
        Additional keyword arguments that are passed and are documented in
        `scipy.integrate.simpson`.

    Returns
    -------
    float
        amorphous short range order parameter

    References
    ----------
    .. [4] Fernandez, F., Paz, S.A., Otero, M., Barraco, D. and Leiva, E.P.,
       2021. Characterization of amorphous Li x Si structures from ReaxFF via
       accelerated exploration of local minima. `Physical Chemistry Chemical
       Physics`, 23(31), pp.16776-16784.
    """
    vol = (4.0 / 3.0) * np.pi * np.power(rcut, 3)

    mask = rdf_x < rcut
    ix = rdf_x[mask]
    iy = 4.0 * np.pi * ix * ix * rdf_y[mask]

    cab = scipy.integrate.simpson(iy, x=ix, **kwargs)

    return np.log(cab / vol)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Functions written in C that are used by exma."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""An extendable molecular analyzer."""

# ============================================================================
# CONSTANTS
# ============================================================================

__author__ = """Francisco Fernandez"""
__email__ = "fernandezfrancisco2195@gmail.com"
__version__ = "0.4.1"


# ============================================================================
# IMPORTS
# ============================================================================

# core
from .core import AtomicSystem  # noqa

# distances
from .distances import pbc_distances  # noqa

# molecular dynamics observables
from .msd import MeanSquareDisplacement  # noqa
from .rdf import RadialDistributionFunction  # noqa
from .cn import CoordinationNumber  # noqa
from .cluster import EffectiveNeighbors, DBSCAN, sro  # noqa

# electrochemistry
from .electrochemistry import (  # noqa
    fractional_volume_change,  # noqa
    formation_energy,  # noqa
    voltage,  # noqa
)  # noqa
from .statistics import BlockAverage  # noqa

# io
from .io import *  # noqa
from .io.reader import XYZ, LAMMPS, read_log_lammps  # noqa
from .io.writer import XYZ, LAMMPS, in_lammps  # noqa
from .io.positions import Positions, spherical_nanoparticle, replicate  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Implementation of Mean Square Displacement."""

# ============================================================================
# IMPORTS
# ============================================================================

import matplotlib.pyplot as plt

import numpy as np

import pandas as pd

from ._mdobservable import MDObservable

# ============================================================================
# CLASSES
# ============================================================================


class MeanSquareDisplacement(MDObservable):
    """Mean Square Displacement (MSD) implementation.

    The MSD is a measure of the deviation of the position of an atom
    with respect to its reference position over time. At each time value
    `t`, the MSD is defined as an ensemble average over all atoms to be
    considered.

    Parameters
    ----------
    ftraj : str
        the string corresponding with the filename with the molecular
        dynamics trajectory

    dt : int or float
        the timestep, how separated the measured frames are from each
        other, in the corresponding time units

    type_e : int or str
        the type of the element for which the msd is going to be calculated

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames

    xyztype : str, default="xyz"
        the string that describes the type of xyz file, to pass to reader, it
        is only necessary if ftraj ends with .xyz extension

    Notes
    -----
    The trajectory must be unwrapped outside the simulation cell, if it
    is wrapped, the image to which it corresponds each atom must be found
    in the trajectory file.
    """

    def __init__(
        self, ftraj, dt, type_e, start=0, stop=-1, step=1, xyztype="xyz"
    ):
        super().__init__(ftraj, start, stop, step, xyztype)

        self.dt = dt
        self.type_e = type_e

    def _local_configure(self, frame):
        """Define the reference frame."""
        # mask of atoms of type e
        self.mask_e_ = frame._mask_type(self.type_e)

        # reference positions
        self.xref_ = frame.x[self.mask_e_]
        self.yref_ = frame.y[self.mask_e_]
        self.zref_ = frame.z[self.mask_e_]
        if frame.ix is not None:
            self.xref_ = self.xref_ + frame.box[0] * frame.ix[self.mask_e_]
            self.yref_ = self.yref_ + frame.box[1] * frame.iy[self.mask_e_]
            self.zref_ = self.zref_ + frame.box[2] * frame.iz[self.mask_e_]

        self.mean_square_displacement = []

    def _accumulate(self, frame):
        """Calculate the msd of a single frame."""
        x = frame.x[self.mask_e_]
        y = frame.y[self.mask_e_]
        z = frame.z[self.mask_e_]

        if frame.ix is not None:
            x = x + frame.box[0] * frame.ix[self.mask_e_]
            y = y + frame.box[1] * frame.iy[self.mask_e_]
            z = z + frame.box[2] * frame.iz[self.mask_e_]

        x = x - self.xref_
        y = y - self.yref_
        z = z - self.zref_

        msd = np.square(x) + np.square(y) + np.square(z)

        self.mean_square_displacement.append(np.mean(msd))

    def calculate(self, box=None):
        """Calculate the MSD.

        Parameters
        ----------
        box : np.array, default=None
            the lenght of the box in each x, y, z direction, required when
            the trajectory is in an xyz file.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` with the time and the msd as columns.
        """
        super()._calculate(box)
        self.df_msd_ = pd.DataFrame(
            {
                "t": self.dt * np.arange(0, self.imed, self.step),
                "msd": np.array(self.mean_square_displacement),
            }
        )

        return self.df_msd_

    def plot(self, ax=None, plot_kws=None):
        """Plot the calculated MSD.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axis, default=None
            current metplotlib axis

        plot_kws : dict, defualt=None
            additional keyword arguments that are passed and are documented
            in `matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.pyplot.Axis
            the axis with the plot
        """
        ax = plt.gca() if ax is None else ax

        plot_kws = {} if plot_kws is None else plot_kws

        ax.set_xlabel("t")
        ax.set_ylabel("msd")
        ax.plot(self.df_msd_["t"], self.df_msd_["msd"], **plot_kws)

        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Implementation of functions of interest in electrochemistry."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import pandas as pd

import scipy.interpolate

# ============================================================================
# FUNCTIONS
# ============================================================================


def fractional_volume_change(df, reference_atoms, reference_volume):
    r"""Fractional volume change (fvc) with respect to an element A.

    The *fvc* points are calculated using a normalization relative to
    the number of atoms in the reference structure and its volume

    .. math::
        fvc = {\frac{N_{ref}}{V_{ref}}} \left(
            {\frac{V_x}{N_x}} - {\frac{V_{ref}}{N_{ref}}}
        \right)

    where :math:`N_{ref}` and :math:`V_{ref}` are the number of atoms
    of type A and the volume of the reference structure, :math:`N_x` and
    :math:`V_x` are the number of atoms of type A and the volume of the
    structure at the :math:`x` concentration.

    Parameters
    ----------
    df : pd.DataFrame
        a `pd.DataFrame` with the values of `x` (concentrations) in
        the first column and then columns with the number of atoms of type A
        in each structure, the corresponding equilibrium values of the volume
        and, optionally, with its respective error.

    reference_atoms : int
        the number of atoms in the reference structure.

    reference_volume : float
        the volume, in the corresponding units, of the reference structue.

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the fractional volume change corresponding
        values to each `x` values and the respective error if it was
        possible to calculate.

    Raises
    ------
    KeyError
        if the number of atoms of type A or the volume are not defined in
        the `pd.DataFrame` as `"natoms_a"` and `"volume"`, respectively.

    Notes
    -----
    The names of the columns in the df must be the following ones:

    `"x"`: `x` values

    `"volume"`: volume

    `"natoms_a"`: number of atoms of type A

    `"err_volume"`: the error of each `"volume"` point, optional.
    """
    if ("natoms_a" not in df) or ("volume" not in df):
        raise KeyError(
            "The number of atoms of type A or the volume of the "
            "structures are not specified in the input DataFrame"
        )

    sfc = reference_volume / reference_atoms
    fvc = [
        ((volume / natoms) - sfc) / sfc
        for natoms, volume in zip(df.natoms_a, df.volume)
    ]

    dffvc_ = {"x": df.x, "fvc": np.array(fvc, dtype=np.float32)}

    if "err_volume" in df:
        errfvc = [
            errvolume / (natoms * sfc)
            for natoms, errvolume in zip(df.natoms_a, df.err_volume)
        ]
        dffvc_["errfvc"] = errfvc = np.array(errfvc, dtype=np.float32)

    return pd.DataFrame(dffvc_)


def formation_energy(df, reference_energy_a, reference_energy_b):
    r"""Ideal approximation to the formation energy (FE).

    .. math::
        E_f(x) = E_x - (x \cdot E_b + E_a)

    where :math:`E_x` is the energy per atom of type A, :math:`x` the
    concentration and :math:`E_a` and :math:`E_b` the cohesive energies
    of A and B bulk materials.

    Parameters
    ----------
    df : pd.DataFrame
        a `pd.DataFrame` with the values of `x` (concentrations) in
        the first column and then columns with the number of atoms of types
        A and the equilibrium values of the potential energy and, optionally,
        with its respective error.

    reference_energy_a : float
        the pure energy of element A in bulk

    reference_energy_b : float
        the pure energy of element B in bulk

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the formation energy corresponding values
        to each `x` values and the respective error if it was possible
        to calculate.

    Notes
    -----
    The names of the columns in the df must be the following ones:

    `"x"`: `x` values

    `"epot"`: potential energy

    `"natoms_a"`: number of atoms of type A

    `"err_epot"`: the error of each `"epot"` point, optional.
    """
    if ("x" not in df) or ("natoms_a" not in df) or ("epot" not in df):
        raise KeyError(
            "The x values, the number of atoms of type A or the "
            "corresponding potential energies are not specified in "
            "the input DataFrame"
        )

    fe = [
        energy / na - (x * reference_energy_b + reference_energy_a)
        for x, na, energy in zip(df.x, df.natoms_a, df.epot)
    ]
    dffe_ = {"x": df.x, "fe": np.array(fe, dtype=np.float32)}

    if "err_epot" in df:
        dffe_["errfe"] = np.asarray(df.err_epot, dtype=np.float32)

    return pd.DataFrame(dffe_)


def voltage(df, nums=50, **kwargs):
    r"""Approximation to the voltage curve.

    The formation energies can be used as an approximation to the Gibbs
    formation energy, then the potential :math:`V` is given by:

    .. math::
        V(x) = -{\frac{dE_f(x)}{dx}}

    Parameters
    ----------
    df : pd.DataFrame
        a `pd.DataFrame` with the values of `x` (concentrations) in
        the first column, the formation energy (`"fe"`) and, optionally,
        the respective error

    nums : int, default=50
        number of points at which to evaluate the spline and its
        derivative

    **kwargs
        additional keyword arguments that are passed and are documented
        in `scipy.interpolate.UnivariateSpline`

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the value of the formation energy after
        the spline and the estimated value to the voltage in function of
        `x`.
    """
    weights = df.errfe if "errfe" in df else None
    spline = scipy.interpolate.UnivariateSpline(
        df.x, df.fe, w=weights, **kwargs
    )
    dspline = spline.derivative()

    xpoints = np.linspace(np.min(df.x), np.max(df.x), nums)
    fe_spline = [spline(x) for x in xpoints]
    voltage = [-dspline(x) for x in xpoints]

    return pd.DataFrame(
        {
            "x": xpoints,
            "fe_spline": np.array(fe_spline, dtype=np.float32),
            "voltage": np.array(voltage, dtype=np.float32),
        }
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Core Atomic System class of exma."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np


# ============================================================================
# CLASSES
# ============================================================================


class AtomicSystem:
    """Define the configurations of the atomic system.

    natoms : int, default=None
        the number of atoms in the atomic system

    box : np.array, default=None
        the box lenght in each direction

    types : np.array, default=None
        the types of the atoms

    idx : np.array, default=None
        the index of the atoms

    q : np.array, default=None
        a property of each atom in the atomic system

    x : np.array, default=None
        the positions of the atoms in the x direction

    y : np.array, default=None
        the positions of the atoms in the y direction

    z : np.array, default=None
        the positions of the atoms in the z direction

    ix : np.array, default=None
        the corresponding image of the positions of the atoms in the x
        direction

    iy : np.array, default=None
        the corresponding image of the positions of the atoms in the y
        direction

    iz : np.array, default=None
        the corresponding image of the positions of the atoms in the z
        direction
    """

    def __init__(
        self,
        natoms=None,
        box=None,
        types=None,
        idx=None,
        q=None,
        x=None,
        y=None,
        z=None,
        ix=None,
        iy=None,
        iz=None,
    ):
        self.natoms = natoms
        self.box = box

        self.idx = idx
        self.types = types

        self.q = q

        self.x = x
        self.y = y
        self.z = z

        self.ix = ix
        self.iy = iy
        self.iz = iz

    def _is_sorted(self):
        """Tells if the array x is sorted (-> True) or not (-> False)."""
        return (np.diff(self.idx) >= 0).all()

    def _sort_frame(self, dontsort=("natoms", "box")):
        """Sort the frame from the sortening of the atoms id."""
        id_argsort = np.argsort(self.idx)

        for key in self.__dict__.keys():
            if (
                key.startswith("_")
                or key in dontsort
                or self.__dict__[key] is None
            ):
                continue

            self.__dict__[key] = self.__dict__[key][id_argsort]

        return self

    def _mask_type(self, atom_type):
        """Get a masked array by an specific type of atom."""
        return self.types == atom_type

    def _natoms_type(self, mask_type):
        """Count the number of atoms of an specific type."""
        return np.count_nonzero(mask_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""exma IO module for interacting with molecular dynamics files."""

# ============================================================================
# IMPORTS
# ============================================================================

import warnings

import numpy as np

from . import reader
from . import writer

# ============================================================================
# FUNCTIONS
# ============================================================================


def xyz2lammpstrj(xyztraj, lammpstrj_name, cell_info, xyzftype="xyz"):
    """Rewrite an xyz file to a lammpstrj file.

    Parameters
    ----------
    xyztraj : str
        the name of the file with the xyz trajectory.

    lammpstrj_name : str
        the name of the file with the lammpstrj trajectory.

    cell_info : dict
        with the `box`, the lenght of the box in each direction, another
        dictionary identified with the `type` key that has within it a
        correspondence between the elements present in xyz file with integer
        identification numbers, e.g. {"Sn": 1, "O": 2}

    xyzftype : str, default="xyz"
        the `ftype` of xyz file.
    """
    with reader.XYZ(xyztraj, xyzftype) as xyz, writer.LAMMPS(
        lammpstrj_name
    ) as lmp:
        try:
            while True:
                xyz_frame = xyz.read_frame()

                xyz_frame.box = cell_info["box"]
                xyz_frame.idx = np.arange(1, xyz_frame.natoms + 1)
                xyz_frame.types = [
                    cell_info["type"][t] for t in xyz_frame.types
                ]

                lmp.write_frame(xyz_frame)

        except EOFError:
            ...


def xyz2inlmp(xyztraj, inlammps_name, cell_info, nframe=-1, xyzftype="xyz"):
    """Write an xyz frame to an input data file of LAMMPS.

    Parameters
    ----------
    xyztraj : str
        the name of the file with the xyz trajectory.

    inlammps_name : str
        the name of the file to write to.

    cell_info : dict
        with the `box`, the lenght of the box in each direction, another
        dictionary identified with the `type` key that has within it a
        correspondence between the elements present in xyz file with integer
        identification numbers, e.g. {"Sn": 1, "O": 2}

    nframe : int, default=-1
        number of the frame to write, by default is -1, that is, the last.

    xyzftype : str, default="xyz"
        the `ftype` of xyz file.
    """
    nframe = np.inf if nframe == -1 else nframe
    with reader.XYZ(xyztraj, xyzftype) as xyz:
        try:
            iframe = 0
            dframe = xyz.read_frame()
            while iframe < nframe:
                xyz_frame = xyz.read_frame()

                iframe += 1
                dframe = xyz_frame

        except EOFError:
            if nframe != np.inf:
                warnings.warn(
                    f"frame {nframe} does not exist in the trajectory file, "
                    f"therefore the last frame ({iframe}) was written."
                )

    dframe.box = cell_info["box"]
    dframe.idx = np.arange(1, dframe.natoms + 1)
    dframe.types = [cell_info["type"][t] for t in dframe.types]
    if "q" in cell_info.keys():
        dframe.q = cell_info["q"]

    writer.in_lammps(inlammps_name, dframe)


def lammpstrj2xyz(lammpstrjtraj, xyz_name, type_info, xyzftype="xyz"):
    """Rewrite a lammpstrj file to an xyz file.

    Parameters
    ----------
    lammpstrjtraj : str
        the name of the file with the lammpstrj trajectory.

    xyz_name : str
        the name of the file with the lammpstrj trajectory.

    type_info : dict
        a correspondence between the elements id present in lammpstrj file
        with str element, e.g. {1: "Sn", 2: "O"}

    xyzftype : str, default="xyz"
        the `ftype` of xyz file.
    """
    with reader.LAMMPS(lammpstrjtraj) as lmp, writer.XYZ(
        xyz_name, xyzftype
    ) as xyz:
        try:
            while True:
                lmp_frame = lmp.read_frame()
                lmp_frame = (
                    lmp_frame._sort_traj()
                    if not lmp_frame._is_sorted()
                    else lmp_frame
                )
                lmp_frame.types = [type_info[t] for t in lmp_frame.types]

                xyz.write_frame(lmp_frame)

        except EOFError:
            ...


def lammpstrj2inlmp(lammpstrjtraj, inlammps_name, nframe=-1):
    """Write a lammpstrj frame to an input data file of LAMMPS.

    Parameters
    ----------
    lammpstrjtraj : str
        the name of the file with the lammpstrj trajectory.

    inlammps_name : str
        the name of the file to write to.

    nframe : int, default=-1
        number of the frame to write, by default is -1, that is, the last.
    """
    nframe = np.inf if nframe == -1 else nframe
    with reader.LAMMPS(lammpstrjtraj) as lmp:
        try:
            iframe = 0
            dframe = lmp.read_frame()
            while iframe < nframe:
                lmp_frame = lmp.read_frame()

                iframe += 1
                dframe = lmp_frame

        except EOFError:
            if nframe != np.inf:
                warnings.warn(
                    f"frame {nframe} does not exist in the trajectory file, "
                    f"therefore the last frame ({iframe}) was written."
                )

    dframe = dframe._sort_traj() if not dframe._is_sorted() else dframe
    writer.in_lammps(inlammps_name, dframe)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Includes class and functions to define atom positions.

It is primarily developed to generate initial structures returned in the
form of a `exma.core.AtomicSystem` for writing to output files with the writer
classes/function that serve as initial conditions for simulations.
"""

# ============================================================================
# IMPORTS
# ============================================================================

import itertools as it

import numpy as np

from ..core import AtomicSystem

# ============================================================================
# CLASSES
# ============================================================================


class Positions:
    r"""Define the positions of the atoms in an orthogonal lattice.

    Parameters
    ----------
    natoms : int
        number of atoms

    box_size : float
        box size in each direction (cubic)

    Notes
    -----
    Note that the density is defined by the input parameters as follow:

    .. math::
        {\rho} = {\frac{N}{L^3}}

    where `N` is the number of atoms and `L` the box lenght in each
    direction.
    """

    def __init__(self, natoms, box_size):

        self.natoms = natoms
        self.box_size = box_size

    def sc(self):
        """Simple-cubic crystal.

        This cell is characterized by having an atom in each of its vertices.

        Returns
        -------
        `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.

        Raises
        ------
        ValueError
            If the number of atoms does not correspond with the number of
            sites that a sc crystal structure has.
        """
        nside = np.cbrt(self.natoms, dtype=np.float32)
        tmp = np.intc(nside)
        if nside % tmp != 0:
            raise ValueError("Number of atoms must be a power of three")

        s_range = range(int(nside))
        positions = list(it.product(s_range, repeat=3))

        positions = np.array(positions, dtype=np.float32)
        positions = np.transpose(positions) * (self.box_size / nside)

        return AtomicSystem(
            natoms=self.natoms,
            box=np.full(3, self.box_size, dtype=np.float32),
            x=positions[0],
            y=positions[1],
            z=positions[2],
        )

    def bcc(self):
        """Body-centered cubic crystal.

        This cell is characterized by having one atom in each of its vertices
        and an additional atom in the center of the cube.

        Returns
        -------
        `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.

        Raises
        ------
        ValueError
            If the number of atoms does not correspond with the number of
            sites that a bcc crystal structure has.
        """
        nside = np.cbrt(self.natoms / 2, dtype=np.float32)
        tmp = np.intc(nside)
        if nside % tmp != 0:
            raise ValueError(
                "Number of atoms must be a power of three multiplied by two"
            )

        s_range = range(int(nside))
        p0 = list(it.product(s_range, repeat=3))

        # bcc lattice vectors: (0, 0, 0) and (0.5, 0.5, 0.5)
        p0 = np.array(p0)
        p1 = p0 + np.full((len(p0), 3), [0.5, 0.5, 0.5])

        positions = np.concatenate((p0, p1), dtype=np.float32)
        positions = np.transpose(positions) * (self.box_size / nside)

        return AtomicSystem(
            natoms=self.natoms,
            box=np.full(3, self.box_size, dtype=np.float32),
            x=positions[0],
            y=positions[1],
            z=positions[2],
        )

    def fcc(self):
        """Face-centered cubic crystal.

        This cell is characterized by having one atom in each of its vertices
        and an additional atom in each of its faces.

        Returns
        -------
        `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.

        Raises
        ------
        ValueError
            If the number of atoms does not correspond with the number of
            sites that a fcc crystal structure has.
        """
        nside = np.cbrt(self.natoms / 4, dtype=np.float32)
        tmp = np.intc(nside)
        if nside % tmp != 0:
            raise ValueError(
                "Number of atoms must be a power of three multiplied by four"
            )

        s_range = range(int(nside))
        p0 = list(it.product(s_range, repeat=3))

        # fcc lattice vectors:
        # (0, 0, 0) (0.5, 0.5, 0) (0.5, 0, 0.5) (0, 0.5, 0.5)
        p0 = np.array(p0)
        p1 = p0 + np.full((len(p0), 3), [0.5, 0.5, 0.0])
        p2 = p0 + np.full((len(p0), 3), [0.5, 0.0, 0.5])
        p3 = p0 + np.full((len(p0), 3), [0.0, 0.5, 0.5])

        positions = np.concatenate((p0, p1, p2, p3), dtype=np.float32)
        positions = np.transpose(positions) * (self.box_size / nside)

        return AtomicSystem(
            natoms=self.natoms,
            box=np.full(3, self.box_size, dtype=np.float32),
            x=positions[0],
            y=positions[1],
            z=positions[2],
        )

    def dc(self):
        """Diamond cubic crystal.

        The typical structure of a diamond, defined by eight sites.

        Returns
        -------
        `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.

        Raises
        ------
        ValueError
            If the number of atoms does not correspond with the number of
            sites that a dc crystal structure has.
        """
        nside = np.cbrt(self.natoms / 8, dtype=np.float32)
        tmp = np.intc(nside)
        if nside % tmp != 0:
            raise ValueError("Number of atoms not valid")

        s_range = range(int(nside))
        p0 = list(it.product(s_range, repeat=3))

        p0 = np.array(p0)
        p1 = p0 + np.full((len(p0), 3), [0.25, 0.75, 0.25])
        p2 = p0 + np.full((len(p0), 3), [0.00, 0.00, 0.50])
        p3 = p0 + np.full((len(p0), 3), [0.25, 0.25, 0.75])
        p4 = p0 + np.full((len(p0), 3), [0.00, 0.50, 0.00])
        p5 = p0 + np.full((len(p0), 3), [0.75, 0.75, 0.75])
        p6 = p0 + np.full((len(p0), 3), [0.50, 0.00, 0.00])
        p7 = p0 + np.full((len(p0), 3), [0.75, 0.25, 0.25])
        p8 = p0 + np.full((len(p0), 3), [0.50, 0.50, 0.50])

        positions = np.concatenate(
            (p1, p2, p3, p4, p5, p6, p7, p8), dtype=np.float32
        )
        positions = np.transpose(positions) * (self.box_size / nside)

        return AtomicSystem(
            natoms=self.natoms,
            box=np.full(3, self.box_size, dtype=np.float32),
            x=positions[0],
            y=positions[1],
            z=positions[2],
        )


# ============================================================================
# FUNCTIONS
# ============================================================================


def spherical_nanoparticle(frame, rcut):
    """Cut a defined structure to give a spherical nanoparticle.

    Parameters
    ----------
    frame : `exma.core.AtomicSystem`
        This have all the information of the configurations of the system.

    rcut : float
        the radius of the nanoparticle.

    Returns
    -------
    `exma.core.AtomicSystem`
        This have all the information of the configurations of the system.

    Notes
    -----
    If `rcut` is greater than half the cell length, the cell will be
    replicated to give a nanoparticle of the desired radius.
    """
    box_size = frame.box
    x = frame.x
    y = frame.y
    z = frame.z

    # scale the positions if they are not in fractions of the box size
    if np.max(frame.x) > 0.5 * box_size[0]:
        x = x / box_size[0]
        y = y / box_size[1]
        z = z / box_size[2]

    n = np.intc(np.ceil(rcut / np.max(box_size)))
    boxes = list(it.product(range(-n, n + 1), repeat=3))

    npty, npx, npy, npz = [], [], [], []
    for box in boxes:
        for i in range(len(x)):
            xx = box_size[0] * (x[i] + box[0])
            yy = box_size[1] * (y[i] + box[1])
            zz = box_size[2] * (z[i] + box[2])

            if np.linalg.norm([xx, yy, zz]) <= rcut:
                npty.append(frame.types[i])

                npx.append(xx)
                npy.append(yy)
                npz.append(zz)

    return AtomicSystem(
        natoms=len(npx),
        types=npty,
        x=np.asarray(npx, dtype=np.float32),
        y=np.asarray(npy, dtype=np.float32),
        z=np.asarray(npz, dtype=np.float32),
    )


def replicate(frame, nrf):
    """Replicate a crystalline system in each direction.

    Parameters
    ----------
    frame : `exma.core.AtomicSystem`
        This have all the information of the configurations of the system.

    nrf : list
        three integers that must be greater than or equal to 1 and
        indicates the replication factor in each x, y, z direction,
        respectively. Value equal to 1 means that only the current cell
        in that direction is considered.

    Returns
    -------
    `exma.core.AtomicSystem`
        This have all the information of the configurations of the system.
    """
    box_size = frame.box
    x = frame.x
    y = frame.y
    z = frame.z

    # scale the positions if they are not in fractions of the box size
    if np.max(frame.x) > 0.5 * box_size[0]:
        x = x / box_size[0]
        y = y / box_size[1]
        z = z / box_size[2]

    boxes = list(it.product(range(np.max(nrf)), repeat=3))
    newx, newy, newz = [], [], []
    for box in boxes:
        if (box[0] >= nrf[0]) or (box[1] >= nrf[1]) or (box[2] >= nrf[2]):
            continue

        for i in range(frame.natoms):
            newx.append(box_size[0] * (x[i] + box[0]))
            newy.append(box_size[1] * (y[i] + box[1]))
            newz.append(box_size[2] * (z[i] + box[2]))

    natoms = len(newx)
    box_size = np.array(
        [nrf[0] * box_size[0], nrf[1] * box_size[1], nrf[2] * box_size[2]]
    )
    atom_type = frame.types * np.prod(nrf)

    return AtomicSystem(
        natoms=natoms,
        box=box_size,
        types=atom_type,
        x=np.asarray(newx, dtype=np.float32),
        y=np.asarray(newy, dtype=np.float32),
        z=np.asarray(newz, dtype=np.float32),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Classes and functions to read MD trayectories."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import pandas as pd

from ._trajectory_rw import TrajectoryReader

# ============================================================================
# CLASSES
# ============================================================================


class XYZ(TrajectoryReader):
    """Class to read xyz files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories in xyz format are

    ftype : str, default="xyz"
        the possible values are `xyz`, `property` and `image`.
        `xyz` if is the usual xyz file. `property` if in the last
        column there is a property. `image` if in the last three columns
        there are the image box of the corresponding atom.

    Raises
    ------
    ValueError
        If xyz file type is not among the possible values
    """

    def __init__(self, filename, ftype="xyz"):
        if ftype not in ["xyz", "property", "image"]:
            raise ValueError("ftype must be 'xyz', 'property' or 'image'")

        super().__init__(filename, ftype)

    def read_frame(self):
        """Read the actual frame of an .xyz file.

        Returns
        -------
        frame : `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.

        Raises
        ------
        EOFError
            If there are no more frames to read
        """
        natoms = self.file_traj_.readline()
        if not natoms:
            raise EOFError("There is no more frames to read")

        natoms = np.intc(natoms)
        self.file_traj_.readline()  # usually a comment in .xyz files

        atom_type = []
        x, y, z = [], [], []
        q = [] if self.ftype == "property" else None
        ix, iy, iz = [], [], [] if self.ftype == "image" else None
        for i in range(natoms):
            xyzline = self.file_traj_.readline().split()

            atom_type.append(xyzline[0])

            x.append(xyzline[1])
            y.append(xyzline[2])
            z.append(xyzline[3])

            if self.ftype == "property":
                q.append(xyzline[4])
            elif self.ftype == "image":
                ix.append(xyzline[4])
                iy.append(xyzline[5])
                iz.append(xyzline[6])

        self.frame.natoms = natoms
        self.frame.types = np.asarray(atom_type, dtype=str)
        self.frame.x = np.asarray(x, dtype=np.float32)
        self.frame.y = np.asarray(y, dtype=np.float32)
        self.frame.z = np.asarray(z, dtype=np.float32)

        self.frame.ix = (
            np.asarray(ix, dtype=np.intc) if self.ftype == "image" else None
        )
        self.frame.iy = (
            np.asarray(iy, dtype=np.intc) if self.ftype == "image" else None
        )
        self.frame.iz = (
            np.asarray(iz, dtype=np.intc) if self.ftype == "image" else None
        )

        self.frame.q = (
            np.asarray(q, dtype=np.float32)
            if self.ftype == "property"
            else None
        )

        return self.frame


class LAMMPS(TrajectoryReader):
    """Class to read lammpstrj files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories of lammps are

    headerint : list, default=["idx", "types", "ix", "iy", "iz"]
        the columns that have int data types, the others are considered floats.
    """

    def __init__(self, filename, headerint=["idx", "types", "ix", "iy", "iz"]):
        super().__init__(filename, None)
        self.headerint = headerint

    def read_frame(self):
        """Read the actual frame of an .lammpstrj file.

        Returns
        -------
        frame : `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.

        Raises
        ------
        EOFError
            If there are no more frames to read
        """
        comment = self.file_traj_.readline()
        if not comment:
            raise EOFError("There is no more frames to read")
        self.file_traj_.readline()
        self.file_traj_.readline()

        natoms = np.intc(self.file_traj_.readline())

        self.file_traj_.readline()

        box_size = []
        for _ in range(3):
            lohi = self.file_traj_.readline().split()
            box_size.append(np.float32(lohi[1]) - np.float32(lohi[0]))
        box = np.array(box_size)

        keys = self.file_traj_.readline().split()[2:]

        # make sure that the keywords id and type of lammps are not used in
        # self.frame.
        keys = list(
            map(
                lambda x: x.replace("id", "idx").replace("type", "types"), keys
            )
        )

        frame = {key: list() for key in keys}
        for _ in range(natoms):
            line = self.file_traj_.readline().split()
            for j, element in enumerate(line):
                frame[keys[j]].append(element)

        # a way to know the type of data (always a np.float32 except when the
        # data corresponds with integers: `id`, `type`, `ix`, `iy`, `iz`)
        dtypes = [
            np.float32 if key not in self.headerint else np.intc
            for key in keys
        ]

        self.frame.natoms = natoms
        self.frame.box = box
        for key, value, dtype in zip(frame.keys(), frame.values(), dtypes):
            self.frame.__dict__[key] = np.array(value, dtype=dtype)

        return self.frame


# ============================================================================
# FUNCTIONS
# ============================================================================


def read_log_lammps(logname="log.lammps"):
    """Read log file of lammps.

    Parameters
    ----------
    logname : str, defalut="log.lammps".
        the name of the file where the thermodynamic info was logged.

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the columns corresponding to the thermodynamic
        info.

    Notes
    -----
    It only works if the first thermo parameter is `Step`.

    """
    with open(logname, "r") as flog:
        # ignore all previous info
        line = flog.readline()
        while line.startswith("Step ") is False:
            line = flog.readline()

        keys = list(line.split())
        thermo = {key: list() for key in keys}

        # append info until subsequent info to be ignored
        line = flog.readline()
        while line.startswith("Loop time") is False:
            for i, element in enumerate(line.split()):
                thermo[keys[i]].append(element)
            line = flog.readline()

        thermo = {
            key: np.array(value, dtype=np.float32)
            for key, value in zip(thermo.keys(), thermo.values())
        }

        return pd.DataFrame(thermo)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Classes and functions to write MD trayectories."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

from ._trajectory_rw import TrajectoryWriter

# ============================================================================
# CLASSES
# ============================================================================


class XYZ(TrajectoryWriter):
    """Class to write xyz files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories in xyz format are going to
        be written

    ftype : str, default="xyz"
        the possible values are `xyz`, `property` and `image`. `xyz`
        if is the usual xyz file. `property` if in the last column there is
        a property. `image` if in the last three columns there are the image
        box of the corresponding atom.

    Raises
    ------
    ValueError
        If xyz file type is not among the possible values
    """

    def __init__(self, filename, ftype="xyz"):
        if ftype not in ["xyz", "property", "image"]:
            raise ValueError("ftype must be 'xyz', 'property' or 'image'")

        super(XYZ, self).__init__(filename, ftype)

    def write_frame(self, frame):
        """Write the actual frame in an .xyz file.

        Parameters
        ----------
        frame : `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.
        """
        self.file_traj_.write(f"{frame.natoms}\n\n")

        for i in range(frame.natoms):
            line = f"{frame.types[i]:s}  "
            line += f"{frame.x[i]:.6e}  "
            line += f"{frame.y[i]:.6e}  "
            line += f"{frame.z[i]:.6e}"

            if self.ftype == "property":
                line += f"  {frame.q[i]:.6e}"
            if self.ftype == "image":
                line += f"  {frame.ix[i]:d}"
                line += f"  {frame.iy[i]:d}"
                line += f"  {frame.iz[i]:d}"

            line += "\n"
            self.file_traj_.write(line)


class LAMMPS(TrajectoryWriter):
    """Class to write lammpstrj files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories of lammps are going to be
        written
    """

    def __init__(self, filename, ftype="custom"):
        super(LAMMPS, self).__init__(filename, ftype)
        self.timestep = 0

    def write_frame(self, frame):
        """Write the actual frame in a .lammpstrj file.

        Parameters
        ----------
        frame : `exma.core.AtomicSystem`
            This have all the information of the configurations of the system.
        """
        self.file_traj_.write("ITEM: TIMESTEP\n")
        self.file_traj_.write(f"{self.timestep:d}\n")
        self.file_traj_.write("ITEM: NUMBER OF ATOMS\n")
        self.file_traj_.write(f"{frame.natoms:d}\n")
        self.file_traj_.write("ITEM: BOX BOUNDS pp pp pp\n")
        for i in range(0, 3):
            self.file_traj_.write(f"0.0\t{frame.box[i]:.6e}\n")

        self.timestep += 1

        traj_header = "ITEM: ATOMS"
        for key in frame.__dict__.keys():
            if key.startswith("_") or key in ("natoms", "box"):
                continue
            elif frame.__dict__[key] is not None:
                traj_header += " " + str(key).replace("idx", "id").replace(
                    "types", "type"
                )
        traj_header += "\n"
        self.file_traj_.write(traj_header)

        for i in range(frame.natoms):
            line = ""
            for key in frame.__dict__.keys():
                if key.startswith("_") or key in ("natoms", "box"):
                    continue
                if frame.__dict__[key] is not None:
                    value = frame.__dict__[key][i]
                    line += (
                        f"{value:.6e}  "
                        if isinstance(value, np.float32)
                        else f"{value}  "
                    )
            line = line.rstrip()
            line += "\n"
            self.file_traj_.write(line)


# ============================================================================
# FUNCTIONS
# ============================================================================


def in_lammps(file_in, frame):
    """Write a frame as an input file for LAMMPS.

    Parameters
    ----------
    file_in : str
        name of the file where you want to write the input info

    frame : `exma.core.AtomicSystem`
        This have all the information of the configurations of the system.
    """
    header = ("natoms", "box")
    with open(file_in, "w") as f_in:

        f_in.write("# the first three lines are comments...\n")

        f_in_header = "# columns in order:"
        for key in frame.__dict__.keys():
            if key.startswith("_") or key in ("natoms", "box"):
                continue
            elif frame.__dict__[key] is not None:
                f_in_header += " " + str(key).replace("idx", "id").replace(
                    "types", "type"
                )
        f_in_header += "\n"
        f_in.write(f_in_header)

        f_in.write("# input file for LAMMPS generated by exma\n")

        f_in.write(f"{frame.natoms:d} atoms\n")

        noat = np.unique(frame.types).size  # number of atom types
        f_in.write(f"{noat:d} atom types\n\n")

        f_in.write(f"0.0 \t {frame.box[0]:.6e} \t xlo xhi\n")
        f_in.write(f"0.0 \t {frame.box[1]:.6e} \t ylo yhi\n")
        f_in.write(f"0.0 \t {frame.box[2]:.6e} \t zlo zhi\n\n")

        f_in.write("Atoms\n\n")

        for i in range(frame.natoms):
            line = ""
            for key in frame.__dict__.keys():
                if key.startswith("_") or key in header:
                    continue
                if frame.__dict__[key] is not None:
                    value = frame.__dict__[key][i]
                    line += (
                        f"{value:.6e}  "
                        if isinstance(value, np.float32)
                        else f"{value}  "
                    )
            line = line.rstrip()
            line += "\n"
            f_in.write(line)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Molecular Dynamics Observable superclass."""

# ============================================================================
# IMPORTS
# ============================================================================

import warnings

import numpy as np

from .io import reader

# ============================================================================
# CLASSES
# ============================================================================


class MDObservable:
    """Class to define the structure of the molecular dynamics observable.

    Parameters
    ----------
    ftraj : str
        the string corresponding with the filename with the molecular
        dynamics trajectory

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames

    xyztype : str, default="xyz"
        the type of the xyzfile.
    """

    def __init__(self, ftraj, start=0, stop=-1, step=1, xyztype="xyz"):
        self.ftraj = ftraj

        self.start = start
        self.stop = stop
        self.step = step

        self.xyztype = xyztype

    def _configure_reader(self):
        """Configure the reader.

        It defines the trajectory reader type (LAMMPS or XYZ) depending on
        the extension (or raises a ValueError) and define the last frame.
        """
        # configure the frame at which stop the calculation
        self.stop = np.inf if self.stop == -1 else self.stop

        # define the trajectory reader
        fextension = self.ftraj.split(".")[-1]
        if fextension not in ["lammpstrj", "xyz"]:
            raise ValueError(
                "The file must have the extension .xyz or .lammpstrj"
            )

        self.traj_ = (
            reader.LAMMPS(self.ftraj)
            if fextension == "lammpstrj"
            else reader.XYZ(self.ftraj, self.xyztype)
        )

    def _local_configure(self, frame):
        """Specific configuration of each observable."""
        raise NotImplementedError("Implemented in child classes.")

    def _accumulate(self, frame):
        """Accumulate the data of the frame."""
        raise NotImplementedError("Implemented in child classes.")

    def _end(self, frame):
        """Finish the calculation and normilize the data."""
        raise NotImplementedError("Implemented in child classes.")

    def _calculate(self, box=None):
        """Observable main loop.

        Parameters
        ----------
        box : np.array, default=None
            the lenght of the box in each x, y, z direction, required when
            the trajectory is in an xyz file.

        Returns
        -------
        None
            but leave everything ready to be able to do self._end()
        """
        self.imed = 0
        self._configure_reader()

        with self.traj_ as traj:
            try:
                for _ in range(self.start):
                    traj.read_frame()

                frame = traj.read_frame()

                # add the box if not in frame
                frame.box = box if box is not None else frame.box

                # sort the traj if is not sorted, this might not be necessary
                # for all observables.
                if frame.idx is not None:
                    frame = (
                        frame._sort_frame()
                        if not frame._is_sorted()
                        else frame
                    )

                self._local_configure(frame)

                nmed = self.stop - self.start
                while self.imed < nmed:
                    if self.imed % self.step == 0:
                        frame.box = box if box is not None else frame.box

                        if frame.idx is not None:
                            frame = (
                                frame._sort_frame()
                                if not frame._is_sorted()
                                else frame
                            )

                        self._accumulate(frame)

                    self.imed += 1
                    frame = traj.read_frame()

            except EOFError:
                if self.stop != np.inf:
                    warnings.warn(
                        f"the trajectory does not read until {self.stop}"
                    )

            finally:
                return None

    def calculate(self):
        """Calculate the observable."""
        raise NotImplementedError("Implemented in child classes.")

    def to_dataframe(self):
        """Convert the results to pandas.DataFrame."""
        raise NotImplementedError("Implemented in child classes.")

    def plot(self):
        """Make a plot of the observable."""
        raise NotImplementedError("Implemented in child classes.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Radial Distribution Function Calculations."""

# ============================================================================
# IMPORTS
# ============================================================================

import ctypes as ct
import os
import pathlib
import sysconfig

import matplotlib.pyplot as plt

import numpy as np

import pandas as pd

from ._mdobservable import MDObservable

# ============================================================================
# CONSTANTS
# ============================================================================

PATH = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))


# ============================================================================
# CLASSES
# ============================================================================


class RadialDistributionFunction(MDObservable):
    r"""Radial Distribution Function (RDF) implementation.

    The RDF is a descriptor of the variation of density of a system in
    function of distance from a reference atom. It gives the probability
    of finding an atom, relative to an ideal gas, at a given distance.

    This microscopic information gives global information about the system,
    for example, if in an RDF plot the peaks are well defined, it means that
    the system is behaving like a solid; on the other hand, if the peaks are
    broadened and decrease in intensity as the distance increases, tending
    to oscillate around 1, it means that the system behaves like a liquid;
    a straight line at 1 is an ideal gas.

    Parameters
    ----------
    ftraj : str
        the string corresponding with the filename with the molecular
        dynamics trajectory

    type_c : int or str
        type of central atoms

    type_i : int or str
        type of interacting atoms

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames

    rmax : float, default=10.0
        the maximum distance at which to calculate g(r), should not be
        greater than half of the shortest lenght of the box if pbc are
        considered

    nbin : int, default=100
        number of bins in the histogram

    pbc : bool, default=True
        True if periodic boundary conditions must be considered, False if
        not.

    Notes
    -----
    The definition of `rmax` and `nbin` defines the `dr`, the width of the
    histogram as

    .. math::
        dr = \frac{rmax}{nbin}

    for example, the default values give `dr=0.1`.
    """

    def __init__(
        self,
        ftraj,
        type_c,
        type_i,
        start=0,
        stop=-1,
        step=1,
        rmax=10.0,
        nbin=100,
        pbc=True,
    ):
        super().__init__(ftraj, start, stop, step)

        self.type_c = type_c
        self.type_i = type_i

        self.rmax = rmax
        self.nbin = nbin
        self.pbc = pbc

        # the width of the bar in the histogram
        self.dgofr_ = self.rmax / self.nbin

    def _local_configure(self, frame):
        """Configure the Radial Distribution Function calculus.

        It receive a frame and define the boundary conditions, initializate the
        volumen, the counter and the ctypes requirements.
        """
        # pbc = True -> 1; False -> 0 in C code
        self.pbc = 1 if self.pbc else 0

        # init volume
        self.volume_ = 0.0

        # g(r) the counter
        self.ngofr_ = 0

        # calculate natoms_c_ and natoms_i_
        self.mask_c_ = frame._mask_type(self.type_c)
        self.mask_i_ = frame._mask_type(self.type_i)
        self.natoms_c_ = frame._natoms_type(self.mask_c_)
        self.natoms_i_ = frame._natoms_type(self.mask_i_)

        # ctypes requirements to interact with C code
        lib_rdf = ct.CDLL(
            str(PATH / "lib" / "lib_rdf")
            + sysconfig.get_config_var("EXT_SUFFIX")
        )
        self.rdf_c_ = lib_rdf.rdf_accumulate
        self.rdf_c_.argtypes = [
            ct.c_int,
            ct.c_int,
            ct.c_void_p,
            ct.c_void_p,
            ct.c_void_p,
            ct.c_int,
            ct.c_float,
            ct.c_float,
            ct.c_int,
            ct.c_void_p,
        ]
        self.gofr_c_ = (ct.c_int * self.nbin)()

    def _accumulate(self, frame):
        """Accumulates the info of each frame."""
        box = frame.box
        self.volume_ += np.prod(box)

        xc, yc, zc = (
            frame.x[self.mask_c_],
            frame.y[self.mask_c_],
            frame.z[self.mask_c_],
        )
        xi, yi, zi = (
            frame.x[self.mask_i_],
            frame.y[self.mask_i_],
            frame.z[self.mask_i_],
        )

        # accomodate the data type of pointers to C code
        box = np.asarray(box, dtype=np.float32)
        box_c = box.ctypes.data_as(ct.POINTER(ct.c_void_p))

        xcentral_c = np.concatenate((xc, yc, zc)).astype(np.float32)
        xcentral_c = xcentral_c.ctypes.data_as(ct.POINTER(ct.c_void_p))

        xinteract_c = np.concatenate((xi, yi, zi)).astype(np.float32)
        xinteract_c = xinteract_c.ctypes.data_as(ct.POINTER(ct.c_void_p))

        # run rdf C accumulation
        self.ngofr_ += 1
        self.rdf_c_(
            self.natoms_c_,
            self.natoms_i_,
            box_c,
            xcentral_c,
            xinteract_c,
            self.pbc,
            self.dgofr_,
            self.rmax,
            self.nbin,
            self.gofr_c_,
        )

    def _end(self):
        """Finish the calculation and normalize the data."""
        r = np.arange(
            self.dgofr_ / 2, self.rmax + self.dgofr_ / 2, self.dgofr_
        )
        gofr = np.asarray(
            np.frombuffer(self.gofr_c_, dtype=np.intc, count=self.nbin)
        )

        volume = self.volume_ / self.ngofr_
        rho = self.natoms_c_ * self.natoms_i_ / volume

        shell_vols = np.power(self.dgofr_, 3) * np.diff(
            np.power(range(self.nbin + 1), 3)
        )
        ideal = 4.0 * np.pi * shell_vols * rho / 3.0

        rdf = gofr / (self.ngofr_ * ideal)

        return r, rdf

    def calculate(self, box=None):
        """Calculate the RDF.

        Parameters
        ----------
        box : np.array, default=None
            the lenght of the box in each x, y, z direction, required when
            the trajectory is in an xyz file.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` with r and g(r) as columns.
        """
        super()._calculate(box)

        r, rdf = self._end()
        self.df_rdf_ = pd.DataFrame({"r": r, "rdf": rdf})

        return self.df_rdf_

    def plot(self, ax=None, plot_kws=None):
        """Plot the calculated RDF.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axis, default=None
            current metplotlib axis

        plot_kws : dict, defualt=None
            additional keyword arguments that are passed and are documented
            in `matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.pyplot.Axis
            the axis with the plot
        """
        ax = plt.gca() if ax is None else ax

        plot_kws = {} if plot_kws is None else plot_kws

        ax.set_xlabel("r")
        ax.set_ylabel("g(r)")
        ax.plot(self.df_rdf_["r"], self.df_rdf_["rdf"], **plot_kws)

        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# =============================================================================
# DOCS
# =============================================================================

"""Installation of exma."""

# =============================================================================
# IMPORTS
# =============================================================================

import os
import pathlib
import sysconfig

from setuptools import find_packages

from distutils.core import Extension, setup  # noqa: I100

# =============================================================================
# CONSTANTS
# =============================================================================

PATH = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))

REQUIREMENTS = [
    "more-itertools",
    "matplotlib",
    "numpy",
    "pandas",
    "scikit-learn",
    "scipy",
]

COMMON = "exma/lib/"
CFLAGS = sysconfig.get_config_var("CFLAGS").split()
CFLAGS += ["-O3", "-fPIC", "-ftree-vectorize", "-march=native"]
C_MODS = [
    Extension(
        COMMON + "lib_" + mod,
        sources=[COMMON + mod + ".c"],
        depends=[COMMON + mod + ".h"],
        extra_compile_args=CFLAGS,
    )
    for mod in ["cn", "pbc_distances", "rdf"]
]

with open(PATH / "exma" / "__init__.py") as fp:
    for line in fp.readlines():
        if line.startswith("__version__ = "):
            VERSION = line.split("=", 1)[-1].replace('"', "").strip()
            break

with open("README.md") as file_readme:
    LONG_DESCRIPTION = file_readme.read()


# =============================================================================
# FUNCTIONS
# =============================================================================

setup(
    name="exma",
    version=VERSION,
    description="extendable molecular dynamics analyzer",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    author="Francisco Fernandez",
    author_email="fernandezfrancisco2195@gmail.com",
    url="https://github.com/fernandezfran/exma",
    license="MIT",
    install_requires=REQUIREMENTS,
    setup_requires=REQUIREMENTS,
    keywords=["exma", "molecular-dynamics", "data-analysis"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: C",
        "Topic :: Scientific/Engineering",
    ],
    ext_modules=C_MODS,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""An estimator of the error of a serie of data.

It is intended to facilitate the estimation of the error of thermodynamic
quantities before using them to calculate electrochemical properties.
"""

# ============================================================================
# IMPORTS
# ============================================================================

import matplotlib.pyplot as plt

import numpy as np

import pandas as pd

# ============================================================================
# FUNCTIONS
# ============================================================================


class BlockAverage:
    """Estimating error method when data are correlated.

    This method offers an easy and efficient way to estimate the error of
    correlated data by a renormalization groups, as described by
    H. Flyvbjerg and H. G. Petersen [1]_.

    Parameters
    ----------
    x : np.array
        where the time series is

    References
    ----------
    .. [1] Flyvbjerg, H. and Petersen, H.G., 1989. Error estimates on averages
       of correlated data. `The Journal of Chemical Physics`, 91(1),
       pp.461-466.
    """

    def __init__(self, x):
        self.x = x

    def calculate(self):
        """Calculate the estimation of the error.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` with `data_size`, `mean`, `var` and `varerr`
            as columns that gives information about the data size changes,
            the mean value of each block, the corresponding variance and the
            error of that variance, respectively.
        """
        data_size, mean, var, varerr = [], [], [], []

        idx = 0
        data_size.append(len(self.x))
        mean.append(np.mean(self.x))
        var.append(np.var(self.x) / (data_size[idx] - 1))
        varerr.append(np.sqrt(2.0 / (data_size[idx] - 1)) * var[idx])

        oldx = self.x
        while np.intc(len(oldx) / 2) > 2:
            newx = np.zeros(np.intc(len(oldx) / 2))

            for k in range(len(newx)):
                newx[k] = 0.5 * (oldx[2 * k - 1] + oldx[2 * k])

            idx += 1
            data_size.append(len(newx))
            mean.append(np.mean(newx))
            var.append(np.var(newx) / (data_size[idx] - 1))
            varerr.append(np.sqrt(2.0 / (data_size[idx] - 1)) * var[idx])

            oldx = newx

        self.df_ = pd.DataFrame(
            data={
                "data_size": np.array(data_size),
                "mean": np.array(mean),
                "var": np.array(var),
                "varerr": np.array(varerr),
            }
        )

        return self.df_

    def plot(self, ax=None, errorbar_kws=None):
        """Flyvbjerg & Petersen plot.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axis, default=None
            current metplotlib axis

        errorbar_kws : dict, default={"marker": "o", "ls": ""}
            additional keyword arguments that are passed and are documented
            in `matplotlib.pyplot.errorbar_kws`.

        Returns
        -------
        matplotlib.pyplot.Axis
            the axis with the plot
        """
        ax = plt.gca() if ax is None else ax

        errorbar_kws = {} if errorbar_kws is None else errorbar_kws
        for key, value in zip(["marker", "ls"], ["o", ""]):
            errorbar_kws.setdefault(key, value)

        ax.set_xlabel("number of blocks operations")
        ax.set_ylabel("block average variance")

        ax.errorbar(
            list(self.df_.index),
            np.asarray(self.df_["var"]),
            yerr=np.asarray(self.df_["varerr"]),
            **errorbar_kws,
        )

        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Coordination Number Calculations."""

# ============================================================================
# IMPORTS
# ============================================================================

import ctypes as ct
import os
import pathlib
import sysconfig

import numpy as np

import pandas as pd

from ._mdobservable import MDObservable

# ============================================================================
# CONSTANTS
# ============================================================================

PATH = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))


# ============================================================================
# CLASSES
# ============================================================================


class CoordinationNumber(MDObservable):
    """Coordination Number (CN) implementation.

    The CN, also called ligancy when we are refearing to the first
    coordination shell, of a central atom is the number of atoms inside the
    empty sphere defined by an outer and an inner cut-off radius.

    Parameters
    ----------
    ftraj : str
        the string corresponding with the filename with the molecular
        dynamics trajectory

    type_c : int or str
        type of central atoms

    type_i : int or str
        type of interacting atoms

    rcut_e : float
        external cut-off radius of the shell

    rcut_i : float, default=0.0
        internal cut-off radius of the shell

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames

    pbc : bool, default=True
        True if periodic boundary conditions must be considered, False if
        not.
    """

    def __init__(
        self,
        ftraj,
        type_c,
        type_i,
        rcut_e,
        rcut_i=0.0,
        start=0,
        stop=-1,
        step=1,
        pbc=True,
    ):
        super().__init__(ftraj, start, stop, step)

        self.type_c = type_c
        self.type_i = type_i

        self.rcut_i = rcut_i
        self.rcut_e = rcut_e

        self.pbc = pbc

    def _local_configure(self, frame):
        """Configure the Coordination Number calculus.

        It receive a frame and define the boundary conditions, sets the counter
        and the ctypes requirements.
        """
        # pbc = True -> 1; False -> 0 in C code
        self.pbc = 1 if self.pbc else 0

        # frame counter
        self.ncn_ = 0

        # calculate masks, natoms_c_ and natoms_i_
        self.mask_c_ = frame._mask_type(self.type_c)
        self.mask_i_ = frame._mask_type(self.type_i)
        self.natoms_c_ = frame._natoms_type(self.mask_c_)
        self.natoms_i_ = frame._natoms_type(self.mask_i_)

        # ctypes requirements to interact with C code
        lib_cn = ct.CDLL(
            str(PATH / "lib" / "lib_cn")
            + sysconfig.get_config_var("EXT_SUFFIX")
        )
        self.cn_c_ = lib_cn.cn_accumulate
        self.cn_c_.argtypes = [
            ct.c_int,
            ct.c_int,
            ct.c_void_p,
            ct.c_void_p,
            ct.c_void_p,
            ct.c_int,
            ct.c_float,
            ct.c_float,
            ct.c_void_p,
        ]
        self.cn_res = (ct.c_int * self.natoms_c_)()

    def _accumulate(self, frame):
        """Accumulates the info of each frame."""
        box = frame.box
        xc, yc, zc = (
            frame.x[self.mask_c_],
            frame.y[self.mask_c_],
            frame.z[self.mask_c_],
        )
        xi, yi, zi = (
            frame.x[self.mask_i_],
            frame.y[self.mask_i_],
            frame.z[self.mask_i_],
        )

        # accomodate the data type of pointers to C code
        box = np.asarray(box, dtype=np.float32)
        box_c = box.ctypes.data_as(ct.POINTER(ct.c_void_p))

        xcentral_c = np.concatenate((xc, yc, zc)).astype(np.float32)
        xcentral_c = xcentral_c.ctypes.data_as(ct.POINTER(ct.c_void_p))

        xinteract_c = np.concatenate((xi, yi, zi)).astype(np.float32)
        xinteract_c = xinteract_c.ctypes.data_as(ct.POINTER(ct.c_void_p))

        # run th cn C accumulation
        self.ncn_ += 1
        self.cn_c_(
            self.natoms_c_,
            self.natoms_i_,
            box_c,
            xcentral_c,
            xinteract_c,
            self.pbc,
            self.rcut_i,
            self.rcut_e,
            self.cn_res,
        )

    def _end(self):
        """Finish the calculation and normalize the data."""
        cn = np.asarray(
            np.frombuffer(self.cn_res, dtype=np.intc, count=self.natoms_c_)
        )
        return cn / self.ncn_

    def calculate(self, box=None):
        """Calculate the CN.

        Parameters
        ----------
        box : np.array, default=None
            the lenght of the box in each x, y, z direction, required when
            the trajectory is in an xyz file.

        Returns
        -------
        tuple
            a tuple with the average cn number and its standard deviation.
        """
        super()._calculate(box)
        self.cn_ = self._end()

        return np.mean(self.cn_), np.std(self.cn_)

    def to_dataframe(self):
        """Convert the results to a pandas.DataFrame.

        A one-column DataFrame where for the central atoms the corresponding
        coordination number averaged over the frames in which it was
        calculated is given and for the interacting atoms (which was not
        calculated) a np.nan.

        Returns
        -------
        df : pd.DataFrame
            DataFrame with the coordination number data in the column "cn".
        """
        j = 0
        cn = []
        for _, value in enumerate(self.mask_c_):
            if value:
                cn.append(self.cn_[j])
                j += 1
            else:
                cn.append(np.nan)

        dict_ = {"cn": np.asarray(cn)}

        return pd.DataFrame(dict_)
